/*
 * Routines linked in exeDSP for new aspect ratio control in media player
 */

#include <stdio.h>

#define ZOOM_UNUSED	 	 0
#define ZOOM_FIT		 1
#define ZOOM_ORIGINAL	 2
#define ZOOM_UNKNOWN	 3
#define ZOOM_FULLSCREEN	 4
#define ZOOM_4_3		 5
#define ZOOM_NON_ANAM	 6
#define ZOOM_16_9		 7
#define ZOOM_ANAM		 8
#define ZOOM_185		 9
#define ZOOM_235		10
#define ZOOM_237		11
#define ZOOM_239		12
#define ZOOM_276		13

#define NUM_ZOOMS		14
#define NUM_SIZE 		(NUM_ZOOMS - ZOOM_4_3)

/* Description messages for menu */
static const unsigned short aspect_msg_fullscreen[] = { 'F','u','l','l',' ','S','c','r','e','e','n',0 };
static const unsigned short aspect_msg_4_3[] = { '4',':','3',0 };
static const unsigned short aspect_msg_non_anam[] = { 'N','o','n',' ','A','n','a','m','o','r','p','h',0 };
static const unsigned short aspect_msg_16_9[] = { '1','6',':','9',0 };
static const unsigned short aspect_msg_anam[] = { 'A','n','a','m','o','r','p','h', 0 };
static const unsigned short aspect_msg_185[] = { '1','.','8','5',':','1', 0 };
static const unsigned short aspect_msg_235[] = { '2','.','3','5',':','1', 0 };
static const unsigned short aspect_msg_237[] = { '2','.','3','7',':','1', 0 };
static const unsigned short aspect_msg_239[] = { '2','.','3','9',':','1', 0 };
static const unsigned short aspect_msg_276[] = { '2','.','7','6',':','1', 0 };

unsigned short *aspect_zoom_messages[NUM_ZOOMS] = {
	NULL,	/* Not valid */
	NULL,	/* Patched in by loader */
	NULL,	/* Patched in by loader */
	NULL,	/* size and position controlled via ptr from GetTDiVideoEnhancer. */
	aspect_msg_fullscreen,
	aspect_msg_4_3,
	aspect_msg_non_anam,
	aspect_msg_16_9,
	aspect_msg_anam,
	aspect_msg_185,
	aspect_msg_235,
	aspect_msg_237,
	aspect_msg_239,
	aspect_msg_276
};

/* Data structures from exeDSP */
extern unsigned char g_TaskManager[5716];
extern unsigned _ZN10CMovieCore17m_PanelResolutionE;
extern unsigned _ZN10CMovieCore20m_CurrentDisplaySizeE;

/* Called functions from exeDSP */
extern int _ZN16CMMPlayerWrapper13GetMmpProfileEiPh(int, unsigned char *);
extern int _ZN16CMMPlayerWrapper13SetMmpProfileEiPh(int, unsigned char *);
extern unsigned _ZN12CTaskManager14GetApplicationE15DTV_APPLICATION(void *, unsigned);
extern void _ZN6PCTask4SendEP7PTEventbb(void *, unsigned, unsigned);

/* Returns pointer to message */

void *aspect_get_zoom_message(void) __attribute__ ((noinline));
void *aspect_get_zoom_message(void)
{
	unsigned char aspect_choosen;

	/* Get current setting */
	if (_ZN16CMMPlayerWrapper13GetMmpProfileEiPh(6,&aspect_choosen) == 0)
	{
		return NULL;
	}

	/* Return string */
	return (aspect_choosen < NUM_ZOOMS) ? aspect_zoom_messages[aspect_choosen] : NULL;
}

/* Handles left/right key in menu */
int aspect_handle_left_right(void *ptr __attribute__ ((unused)), unsigned key) __attribute__ ((noinline));
int aspect_handle_left_right(void *ptr __attribute__ ((unused)), unsigned key)
{
	unsigned char aspect_choosen;
	unsigned an_address;
	struct {
		unsigned a;
		unsigned b;
		unsigned c;
		unsigned d;
		unsigned e;
		unsigned f;
	} event;

	/* Get current setting */
	if (_ZN16CMMPlayerWrapper13GetMmpProfileEiPh(6,&aspect_choosen) == 0)
	{
		return 0;
	}

	/* Process key */
	if (key == 5) /* right */
	{
		aspect_choosen++;
		if (aspect_choosen == ZOOM_UNKNOWN)
		{
			aspect_choosen++;
		}
		else if (aspect_choosen >= NUM_ZOOMS)
		{
			aspect_choosen = ZOOM_FIT;
		}
	}
	else
	{
		aspect_choosen--;
		if (aspect_choosen == ZOOM_UNUSED)
		{
			aspect_choosen = NUM_ZOOMS - 1;
		}
		else if (aspect_choosen == ZOOM_UNKNOWN)
		{
			aspect_choosen = ZOOM_ORIGINAL;
		}
	}

	/* Store new setting */
	if (_ZN16CMMPlayerWrapper13SetMmpProfileEiPh(6,&aspect_choosen) == 0)
	{
		return 0;
	}

	/* Update used display settings */
	an_address = _ZN12CTaskManager14GetApplicationE15DTV_APPLICATION(g_TaskManager, 32);

	event.b = (an_address) ? (an_address + 0x144C) : 0;
	event.c = 0x491;
	event.d = 0;
	event.e = 0;

	_ZN6PCTask4SendEP7PTEventbb(&event, 0, 0);

	return 1;
}

/* Calculates the display size
 * Settings ZOOM_FIT, ZOOM_UNKNOWN and ZOOM_FULLSCREEN already handled in
 * original routine */

struct s_window_settings {
	unsigned offset_x;
	unsigned offset_y;
	unsigned size_x;
	unsigned size_y;
};

int aspect_calculate_display_size(struct s_window_settings *settings, unsigned movie_x, unsigned movie_y) __attribute__ ((noinline));
int aspect_calculate_display_size(struct s_window_settings *settings, unsigned movie_x, unsigned movie_y)
{
	unsigned panelresolution_x;
	unsigned panelresolution_y;
	unsigned factor1;
	unsigned factor2;
	unsigned panelindex;
	unsigned windowindex;
	static const unsigned aspect_x_sizes[2][NUM_SIZE] = {
		/* Panel-Resolution 1366 * 768 */
		{ 1024, 1050, 1366, 1366, 1366, 1366, 1366, 1366, 1366 },
		/* Panel-Resolution 1920 * 1080 */
		{ 1440, 1476, 1920, 1920, 1920, 1920, 1920, 1920, 1920 }
	};
	static const unsigned aspect_y_sizes[2][NUM_SIZE] = {
		/* Panel-Resolution 1366 * 768 */
		{  768,  768,  768,  750,  738,  580,  576,  570,  494 },
		/* Panel-Resolution 1920 * 1080 */
		{ 1080, 1080, 1080, 1054, 1038,  816,  810,  804,  696 }
	};

	if (_ZN10CMovieCore17m_PanelResolutionE == 1)
	{
		panelresolution_x = 1366;
		panelresolution_y = 768;
		panelindex = 0;
	}
	else
	{
		panelresolution_x = 1920;
		panelresolution_y = 1080;
		panelindex = 1;
	}

	if (_ZN10CMovieCore20m_CurrentDisplaySizeE == ZOOM_ORIGINAL)
	{
		/* Has movie a higher resolution than the panel? */
		if ( (movie_x > panelresolution_x) ||
			 (movie_y > panelresolution_y) )
		{
			factor1 = (panelresolution_x << 16) / movie_x;
			factor2 = (panelresolution_y << 16) / movie_y;
			if (factor2 < factor1)
			{
				factor1 = factor2;
			}
			settings->size_x = (movie_x * factor1) >> 16;
			settings->size_y = (movie_y * factor1) >> 16;
		}
		else
		{
			settings->size_x = movie_x;
			settings->size_y = movie_y;
		}
	}
	else
	{
		windowindex = _ZN10CMovieCore20m_CurrentDisplaySizeE - ZOOM_4_3;
		if (windowindex >= NUM_SIZE)
		{
			settings->size_x = panelresolution_x;
			settings->size_y = panelresolution_y;
		}
		else
		{
			settings->size_x = aspect_x_sizes[panelindex][windowindex];
			settings->size_y = aspect_y_sizes[panelindex][windowindex];
		}
	}

	settings->offset_x = (panelresolution_x - settings->size_x) / 2;
	settings->offset_y = (panelresolution_y - settings->size_y) / 2;

	/* Restore original R0 value */
	return _ZN10CMovieCore20m_CurrentDisplaySizeE;
}
