/*
 * - mounts /dev/pts and starts telnetd
 * - fixes zooming of media player
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/mount.h>
#include <sys/vfs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/mman.h>

extern char **environ;
extern int Game_Main(const char *path, const char *udn);

static void start_telnet()
{
	struct statfs fs;
	DIR *dir;
	struct dirent *ent;
	unsigned telnetd_running;
	const char *c;
	unsigned isnum;
	int fd;
	char filename[24];
	pid_t p;
	static char *telnetd_args[2] = {
		"telnetd",
		NULL
	};

	/* Filesystem already mounted? */
	if (statfs("/dev/pts",&fs) < 0)
	{
		return;
	}

	/* Mount it */
	if ( (fs.f_type != 0x1cd1) && (mount("devpts","/dev/pts","devpts",0,"") < 0) )
	{
		return;
	}

	/* Is telnetd running? */
	telnetd_running = 0;

	if ( !(dir = opendir("/proc")) )
	{
		return;
	}

	while (!telnetd_running && (ent = readdir(dir)))
	{
		/* A process id? */
		isnum = 1;
		c = ent->d_name;
		while (isnum && *c)
		{
			isnum = (*c >= '0') && (*c <= '9');
			c++;
		}
		if (isnum)
		{
			sprintf(filename,"/proc/%s/cmdline",ent->d_name);
			if ( (fd = open(filename,O_RDONLY)) >= 0)
			{
				if ( (read(fd,filename,sizeof(filename)) == 8) && !strcmp(filename,"telnetd"))
				{
					telnetd_running = 1;
				}
				close(fd);
			}
		}
	}
	closedir(dir);

	if (telnetd_running)
	{
		return;
	}

	/* Start telnetd */
	chdir("/");
	if ( (p = vfork()) > 0)
	{
		waitpid(p,NULL,0);
		return;
	}
	else if (p < 0)
	{
		return;
	}

	/* Child-Code. */
	execve("/bin/busybox",telnetd_args,environ);
	_exit(0);
}

extern void* _ZN16CResourceManager10GetWStringEi(void *, unsigned);
extern unsigned char g_ToolResMgr[80];

static void patch_aspect(const char *path)
{
	unsigned short **aspect_zoom_messages;
	void *handle;
	void *myself;
	unsigned *original1;
	unsigned *patch1;
	unsigned *original2;
	unsigned *patch2;
	unsigned *original3;
	unsigned *patch3;
	unsigned page1;
	unsigned page2;
	unsigned page3;
	char filename[64];
#define PATCH1_OFFSET	0x88
	static const unsigned pattern1[8] =
	{
		0xE3500002,
		0x0A00006C,
		0xE3500003,
		0x0A000060,
		0xE3500001,
		0x0A00002C,
		0xE3A03000,
		0xE50B7098
	};
	static const unsigned  patchcode1[6] =
	{
		0xE3500001,		/* CMP R0, #1		; FIT */
		0x0A000030,		/* BEQ fit_code */
		0xE3500003,		/* CMP R0, #3		; VideoEnhancerMode */
		0x0A000060,		/* BEQ videoenhander_code */
		0xE3500004,		/* CMP R0, #4		; Full Screen */
		0x1A000068		/* BNE patch_area2 */
	};
#define PATCH2_OFFSET	0x244
	static const unsigned pattern2[8] =
	{
		0xE1560008,
		0x20683006,
		0x31A030A6,
		0x21A030A3,
		0x350B309C,
		0x250B309C,
		0xE06A3007,
		0xE50BA098
	};
	static const unsigned patchcode2[5] =
	{
		0xE24B00A0,		/* SUB R0, R11, #0xA0	; Pointer to size/position vars */
		0xE1A0100A,		/* MOV R1, R10			; X-resolution movie */
		0xE1A02008,		/* MOV R2, R8			; Y-resolution movie */
		0xE24FEF69,		/* SUB LR, PC, #0x1A4	; Return address  */
		0xE51FF004		/* LDR PC,=ROUTINE		; Jump to c function */
	};

	/* Get addresses of functions to patch */
	if ( (myself = dlopen(NULL,RTLD_NOW | RTLD_GLOBAL)) == NULL )
	{
		return;
	}

	original1 = dlsym(myself,"_ZNK23CToolMmbDisplaySizeItem19GetToolItemTextDataEv");
	original2 = dlsym(myself,"_ZNK23CToolMmbDisplaySizeItem17PressLeftRightKeyEi");
	original3 = dlsym(myself,"_ZN10CMovieCore20CalculateDisplaySizeEN7TDiMpeg19TDAVDec_VideoInfo_tE");
	dlclose(myself);

	if ( (original1 == NULL) ||
		 (original2 == NULL) ||
		 (original3 == NULL) )
	{
		return;
	}

	/* Verify original code */
	if ( memcmp(pattern1, original3 + (PATCH1_OFFSET / 4),sizeof(pattern1)) ||
		 memcmp(pattern2, original3 + (PATCH2_OFFSET / 4),sizeof(pattern2)) )
	{
		return;
	}

	/* Load aspect library */
	strcpy(filename,path);
	strcat(filename,"aspect.so");
	if ( (handle = dlopen(filename,RTLD_NOW | RTLD_LOCAL | RTLD_NODELETE)) == NULL )
	{
		return;
	}

	/* Fill in original messages */
	if ( (aspect_zoom_messages = dlsym(handle,"aspect_zoom_messages")) == NULL )
	{
		dlclose(handle);
		return;
	}

	aspect_zoom_messages[1] = _ZN16CResourceManager10GetWStringEi(g_ToolResMgr,0x53);
	aspect_zoom_messages[2] = _ZN16CResourceManager10GetWStringEi(g_ToolResMgr,0x54);

	/* Get addresses of replacement/patch functions */
	patch1 = dlsym(handle,"aspect_get_zoom_message");
	patch2 = dlsym(handle,"aspect_handle_left_right");
	patch3 = dlsym(handle,"aspect_calculate_display_size");
	dlclose(handle);

	if ( (patch1 == NULL) ||
		 (patch2 == NULL) ||
		 (patch3 == NULL) )
	{
		return;
	}

	/* Patch */
	page1 = ((unsigned) original1) & ~0x0FFF;
	page2 = ((unsigned) original2) & ~0x0FFF;
	page3 = ((unsigned) original3) & ~0x0FFF;

	mprotect((void *) page1, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	original1[0] = 0xE51FF004;
	original1[1] = (unsigned) patch1;

	if (page1 != page2)
	{
		mprotect((void *) page1, 8192, PROT_READ | PROT_EXEC);
		mprotect((void *) page2, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	}
	original2[0] = 0xE51FF004;
	original2[1] = (unsigned) patch2;

	if (page2 != page3)
	{
		mprotect((void *) page2, 8192, PROT_READ | PROT_EXEC);
		mprotect((void *) page3, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	}
	memcpy(original3 + (PATCH1_OFFSET / 4),patchcode1,sizeof(patchcode1));
	memcpy(original3 + (PATCH2_OFFSET / 4),patchcode2,sizeof(patchcode2));
	original3[(PATCH2_OFFSET + sizeof(patchcode2)) / 4] = (unsigned) patch3;

	mprotect((void *) page3, 8192, PROT_READ | PROT_EXEC);
}

int Game_Main(const char  *path, const char *udn __attribute__ ((unused)))
{
	start_telnet();
	patch_aspect(path);

	return 0;
}
