/*
 * - mounts /dev/pts and starts telnetd
 * - fixes zooming of media player
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/mman.h>

/* #define DEBUG */

#ifdef DEBUG
FILE *msg;
#endif

static void cacheflush(void *start, size_t size)
{
   /* For details, please see the Linux sources, files
    * arch/arm/kernel/traps.c, arch/arm/kernel/entry-common.S and
    * asm-arm/unistd.h
    *
    * This syscall is supported by all 2.4 and 2.6 Linux kernels.  2.2 linux
    * kernels got first support with version 2.2.18.
    */

   asm("mov r0, %0\n"
       "mov r1, %1\n"
       "mov r2, #0\n"
       /* EABI or Thumb syscall: syscall number passed in 'r7' (syscall base
	* number is 0x0).  Note that Thumb and EABI are generally not the
	* same.  It just happens that the simple cacheflush syscall doesn't
	* expose any of differences in calling conventions.
        */
       "mov r7, #0xf0000\n"
       "add r7, #0x02\n"
       "swi #0\n" ::
       "g"(start), "g"((long)start+size) :
       "r0", "r1", "r2", "r7");
}

extern void* _ZN16CResourceManager10GetWStringEi(void *, unsigned);
extern unsigned char g_ToolResMgr[80];

static void patch_aspect(const char *path)
{
	unsigned short **aspect_zoom_messages;
	void *handle;
	void *myself;
	unsigned *original1;
	unsigned *patch1;
	unsigned *original2;
	unsigned *patch2;
	unsigned *original3;
	unsigned *patch3;
	unsigned page1;
	unsigned page2;
	unsigned page3;
	char filename[64];
#define PATCH1_OFFSET	0x7C
	static const unsigned pattern1[8] =
	{
		0xE3500002,
		0x0A00006C,
		0xE3500003,
		0x0A000061,
		0xE3500001,
		0x0A00002D,
		0xE3A03000,
		0xE58D705C
	};
	static const unsigned  patchcode1[6] =
	{
		0xE3500001,		/* CMP R0, #1		; FIT */
		0x0A000031,		/* BEQ fit_code */
		0xE3500003,		/* CMP R0, #3		; VideoEnhancerMode */
		0x0A000061,		/* BEQ videoenhander_code */
		0xE3500004,		/* CMP R0, #4		; Full Screen */
		0x1A000068		/* BNE patch_area2 */
	};
#define PATCH2_OFFSET	0x238
	static const unsigned pattern2[8] =
	{
		0xE1560008,
		0x20683006,
		0x31A030A6,
		0x21A030A3,
		0x358D3058,
		0x258D3058,
		0xE06A3007,
		0xE58DA05C
	};
	static const unsigned patchcode2[5] =
	{
		0xE28D0060,		/* ADD R0, SP, #0x60	; Pointer to size/position vars */
		0xE1A0100A,		/* MOV R1, R10			; X-resolution movie */
		0xE1A02008,		/* MOV R2, R8			; Y-resolution movie */
		0xE24FEF69,		/* SUB LR, PC, #0x1A4	; Return address  */
		0xE51FF004		/* LDR PC,=ROUTINE		; Jump to c function */
	};

	/* Get addresses of functions to patch */
	if ( (myself = dlopen(NULL,RTLD_NOW | RTLD_GLOBAL)) == NULL )
	{
#ifdef DEBUG
		fputs("aspect: cannot dlopen myself\n",msg);
#endif
		return;
	}

	original1 = dlsym(myself,"_ZNK23CToolMmbDisplaySizeItem19GetToolItemTextDataEv");
	original2 = dlsym(myself,"_ZNK23CToolMmbDisplaySizeItem17PressLeftRightKeyEi");
	original3 = dlsym(myself,"_ZN10CMovieCore20CalculateDisplaySizeEN7TDiMpeg19TDAVDec_VideoInfo_tE");
	dlclose(myself);

	if ( (original1 == NULL) ||
		 (original2 == NULL) ||
		 (original3 == NULL) )
	{
#ifdef DEBUG
		fputs("aspect: cannot locate original functions\n",msg);
#endif
		return;
	}

	/* Verify original code */
	if ( memcmp(pattern1, original3 + (PATCH1_OFFSET / 4),sizeof(pattern1)) ||
		 memcmp(pattern2, original3 + (PATCH2_OFFSET / 4),sizeof(pattern2)) )
	{
#ifdef DEBUG
		fputs("aspect: original function code doesn't match\n",msg);
#endif
		return;
	}

	/* Load aspect library */
	strcpy(filename,path);
	strcat(filename,"aspect.so");
	if ( (handle = dlopen(filename,RTLD_NOW | RTLD_LOCAL | RTLD_NODELETE)) == NULL )
	{
#ifdef DEBUG
		fputs("aspect: cannot load aspect.so\n",msg);
#endif
		return;
	}

	/* Fill in original messages */
	if ( (aspect_zoom_messages = dlsym(handle,"aspect_zoom_messages")) == NULL )
	{
#ifdef DEBUG
		fputs("aspect: cannot locate messages in aspect.so\n",msg);
#endif
		dlclose(handle);
		return;
	}

	aspect_zoom_messages[1] = _ZN16CResourceManager10GetWStringEi(g_ToolResMgr,0x53);
	aspect_zoom_messages[2] = _ZN16CResourceManager10GetWStringEi(g_ToolResMgr,0x54);

	/* Get addresses of replacement/patch functions */
	patch1 = dlsym(handle,"aspect_get_zoom_message");
	patch2 = dlsym(handle,"aspect_handle_left_right");
	patch3 = dlsym(handle,"aspect_calculate_display_size");
	dlclose(handle);

	if ( (patch1 == NULL) ||
		 (patch2 == NULL) ||
		 (patch3 == NULL) )
	{
#ifdef DEBUG
		fputs("aspect: cannot locate new functions\n",msg);
#endif
		return;
	}

	/* Patch */
	page1 = ((unsigned) original1) & ~0x0FFF;
	page2 = ((unsigned) original2) & ~0x0FFF;
	page3 = ((unsigned) original3) & ~0x0FFF;

	mprotect((void *) page1, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	original1[0] = 0xE51FF004;
	original1[1] = (unsigned) patch1;

	if (page1 != page2)
	{
		mprotect((void *) page1, 8192, PROT_READ | PROT_EXEC);
		mprotect((void *) page2, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	}
	original2[0] = 0xE51FF004;
	original2[1] = (unsigned) patch2;

	if (page2 != page3)
	{
		mprotect((void *) page2, 8192, PROT_READ | PROT_EXEC);
		mprotect((void *) page3, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	}
	memcpy(original3 + (PATCH1_OFFSET / 4),patchcode1,sizeof(patchcode1));
	memcpy(original3 + (PATCH2_OFFSET / 4),patchcode2,sizeof(patchcode2));
	original3[(PATCH2_OFFSET + sizeof(patchcode2)) / 4] = (unsigned) patch3;

	mprotect((void *) page3, 8192, PROT_READ | PROT_EXEC);

	cacheflush(original1,2);
	cacheflush(original2,2);
	cacheflush(original3 + (PATCH1_OFFSET / 4),sizeof(patchcode1) / 4);
	cacheflush(original3 + (PATCH2_OFFSET / 4),sizeof(patchcode2) / 4);

#ifdef DEBUG
		fputs("aspect: patch applied\n",msg);
#endif
}

int Game_Main(const char  *path, const char *udn __attribute__ ((unused)))
{
#ifdef DEBUG
	msg = fopen("/dev/kmsg","a+");
#endif
	patch_aspect(path);
#ifdef DEBUG
	fclose(msg);
#endif
	return 1;
}
